#include <TimeLib.h>
#include <time.h>
#include <ESP8266WiFi.h>
#include <WiFiUdp.h>
#include <Adafruit_NeoPixel.h>
const char ssid[] = "Your_SSID";  //  your network SSID (name)
const char pass[] = "Password";       // your network password

int DST = 0; // this will add the offset of one hour during the summer
time_t prevDisplay = 0; // when the digital clock was displayed

#define PIN 2   // GPIO2 = D4 on the board - all the output data for the neopixels on this pin - green wire
//   NEO_KHZ800  800 KHz bitstream (most NeoPixel products w/WS2812 LEDs)
//   NEO_GRB     Pixels are wired for GRB bitstream (most NeoPixel products)
Adafruit_NeoPixel strip = Adafruit_NeoPixel(60, PIN, NEO_GRB + NEO_KHZ800);
const uint32_t red = strip.Color(30, 0, 0);
const uint32_t green = strip.Color(0, 30, 0);
const uint32_t blue = strip.Color(0, 0, 30);
const uint32_t amber = strip.Color(30, 5, 0);
const uint32_t bright_white = strip.Color(30, 30, 30);
const uint32_t dim_white = strip.Color(10, 10, 10);

// NTP Servers:
static const char ntpServerName[] = "nz.pool.ntp.org";
const int timeZone = +12;     // Wellington, New Zealand (GMT+12)

WiFiUDP Udp;
unsigned int localPort = 8888;  // local port to listen for UDP packets

time_t getNtpTime();
void sendNTPpacket(IPAddress &address);

void colorWipe(uint32_t c, uint8_t wait) {
  for(uint16_t i=30; i<60; i++) {
    strip.setPixelColor(i, c);
    strip.show();
    delay(wait);
  }
  for(uint16_t i=0; i<30; i++) {
    strip.setPixelColor(i, c);
    strip.show();
    delay(wait);
  }
}
void all_neo_pixels_off(){
  for (uint16_t i=30; i<60; i++){
    strip.setPixelColor(i,0);
  }
  for (uint16_t i=0; i<30; i++){
    strip.setPixelColor(i,0);
  }
  strip.show(); // all pixels to 'off' 
}

void quarter_past_chime(){
  all_neo_pixels_off();
  for (uint16_t i=30; i<45; i++){
    strip.setPixelColor(i,bright_white);
    strip.show();
    delay(100);
  }
  // flash all pixels on and off three times to show quarters
  for (uint8_t triple=0; triple<4; triple++){
    for (uint8_t i=30; i<45; i++){
      strip.setPixelColor(i,0);
    }
    strip.show();
    delay(300);
    for (uint8_t i=30; i<45; i++){
      strip.setPixelColor(i,bright_white);
    }
    strip.show();
    delay(300);
  }
}

void half_past_chime(){
  all_neo_pixels_off();
  for (uint16_t i=30; i<60; i++){
    strip.setPixelColor(i,bright_white);
    strip.show();
    delay(100);
  }
  // flash all pixels on and off three times to show quarters
  for (uint8_t triple=0; triple<4; triple++){
    for (uint8_t i=30; i<60; i++){
      strip.setPixelColor(i,0);
    }
    strip.show();
    delay(300);
    for (uint8_t i=30; i<60; i++){
      strip.setPixelColor(i,bright_white);
    }
    strip.show();
    delay(300);
  }
}

void quarter_to_chime(){
  all_neo_pixels_off();
  for (uint16_t i=30; i<60; i++){
    strip.setPixelColor(i,bright_white);
    strip.show();
    delay(100);
  }
  for (uint16_t i=0; i<15; i++){
    strip.setPixelColor(i,bright_white);
    strip.show();
    delay(100);
  }
  // flash all pixels on and off three times to show quarters
  for (uint8_t triple=0; triple<4; triple++){
    for (uint8_t i=30; i<60; i++){
      strip.setPixelColor(i,0);  
    }
    for (uint8_t i=0; i<15; i++){
      strip.setPixelColor(i,0);
    }
    strip.show();
    delay(300);
    for (uint8_t i=30; i<60; i++){
      strip.setPixelColor(i,bright_white);
    }
    for (uint8_t i=0; i<15; i++){
      strip.setPixelColor(i,bright_white);
    }
    strip.show();
    delay(300);
  }
}

void o_clock_chime(){
  all_neo_pixels_off();
  colorWipe(bright_white, 100);
  // flash all pixels on and off three times to show quarters
  for (uint8_t triple=0; triple<4; triple++){
    for (uint8_t i=30; i<60; i++){
      strip.setPixelColor(i,0);  
    }
    for (uint8_t i=0; i<30; i++){
      strip.setPixelColor(i,0);
    }
    strip.show();
    delay(300);
    for (uint8_t i=30; i<60; i++){
      strip.setPixelColor(i,bright_white);
    }
    for (uint8_t i=0; i<30; i++){
      strip.setPixelColor(i,bright_white);
    }
    strip.show();
    delay(300);
  } 
}

void rainbow(uint8_t wait) {
  uint16_t i, j;

  for(j=0; j<256; j++) {
    for(i=0; i<strip.numPixels(); i++) {
      strip.setPixelColor(i, Wheel((i+j) & 255));
    }
    strip.show();
    delay(wait);
  }
}

// Slightly different, this makes the rainbow equally distributed throughout
void rainbowCycle(uint8_t wait) {
  uint16_t i, j;
  for(j=0; j<256*5; j++) { // 5 cycles of all colors on wheel
    for(i=0; i< strip.numPixels(); i++) {
      strip.setPixelColor(i, Wheel(((i * 256 / strip.numPixels()) + j) & 255));
    }
    strip.show();
    delay(wait);
  }
}

//Theatre-style crawling lights.
void theaterChase(uint32_t c, uint8_t wait) {
  for (int j=0; j<10; j++) {  //do 10 cycles of chasing
    for (int q=0; q < 3; q++) {
      for (uint16_t i=0; i < strip.numPixels(); i=i+3) {
        strip.setPixelColor(i+q, c);    //turn every third pixel on
      }
      strip.show();

      delay(wait);

      for (uint16_t i=0; i < strip.numPixels(); i=i+3) {
        strip.setPixelColor(i+q, 0);        //turn every third pixel off
      }
    }
  }
}

//Theatre-style crawling lights with rainbow effect
void theaterChaseRainbow(uint8_t wait) {
  for (int j=0; j < 256; j++) {     // cycle all 256 colors in the wheel
    for (int q=0; q < 3; q++) {
      for (uint16_t i=0; i < strip.numPixels(); i=i+3) {
        strip.setPixelColor(i+q, Wheel( (i+j) % 255));    //turn every third pixel on
      }
      strip.show();
      delay(wait);
      for (uint16_t i=0; i < strip.numPixels(); i=i+3) {
        strip.setPixelColor(i+q, 0);        //turn every third pixel off
      }
    }
  }
}

// Input a value 0 to 255 to get a color value.
// The colours are a transition r - g - b - back to r.
uint32_t Wheel(byte WheelPos) {
  WheelPos = 255 - WheelPos;
  if(WheelPos < 85) {
    return strip.Color(255 - WheelPos * 3, 0, WheelPos * 3);
  }
  if(WheelPos < 170) {
    WheelPos -= 85;
    return strip.Color(0, WheelPos * 3, 255 - WheelPos * 3);
  }
  WheelPos -= 170;
  return strip.Color(WheelPos * 3, 255 - WheelPos * 3, 0);
}

// Calculate the day of the week using Sakamoto's Algorithm
String find_day_of_the_week(int d, int m,int y){
int t[]={0,3,2,5,0,3,5,1,4,6,2,4};
y = y - (m < 3);
int dow = (y + y/4 - y/100 + y/400 + t[m-1] + d) % 7;
String day_of_week = "";
switch (dow){
  case 0:
      day_of_week = "Sunday";
      break;
  case 1:
      day_of_week = "Monday";
      break;
  case 2:
      day_of_week = "Tuesday";
      break;
  case 3:
      day_of_week = "Wednesday";
      break;
  case 4:
      day_of_week = "Thursday";
      break;
  case 5:
      day_of_week = "Friday";
      break;
  case 6:
      day_of_week = "Saturday";
      break;
}
return day_of_week;
}

//Function extracted from the library http://playground.arduino.cc/Code/Time
#define SECS_PER_MIN  (60UL)
#define SECS_PER_HOUR (3600UL)
#define SECS_PER_DAY  (SECS_PER_HOUR * 24UL)

// leap year calulator expects year argument as years offset from 1970
#define LEAP_YEAR(Y)     ( ((1970+Y)>0) && !((1970+Y)%4) && ( ((1970+Y)%100) || !((1970+Y)%400) ) )
static  const uint8_t monthDays[]={31,28,31,30,31,30,31,31,30,31,30,31}; // API starts months from 1, this array starts from 0

long makeTime(int hr,int min,int sec,int dy, int mnth, int yr){   
  // assemble time elements into time_t 
  // note year argument is offset from 1970 (see macros in time.h to convert to other formats)
  // previous version used full four digit year (or digits since 2000),i.e. 2009 was 2009 or 9
  
  // year can be given as full four digit year or two digts (2010 or 10 for 2010);  
  //it is converted to years since 1970
  if( yr > 99)
      yr = yr - 1970;
  else
      yr += 30;  

  int i;
  uint32_t seconds;

  // seconds from 1970 till 1 jan 00:00:00 of the given year
  seconds= yr*(SECS_PER_DAY * 365);
  for (i = 0; i < yr; i++) {
    if (LEAP_YEAR(i)) {
      seconds +=  SECS_PER_DAY;   // add extra days for leap years
    }
  }
  
  // add days for this year, months start from 1
  for (i = 1; i < mnth; i++) {
    if ( (i == 2) && LEAP_YEAR(yr)) { 
      seconds += SECS_PER_DAY * 29;
    } else {
      seconds += SECS_PER_DAY * monthDays[i-1];  //monthDay array starts from 0
    }
  }
  seconds+= (dy-1) * SECS_PER_DAY;
  seconds+= hr * SECS_PER_HOUR;
  seconds+= min * SECS_PER_MIN;
  seconds+= sec;
  return (long)seconds; 
}

void setup()
{
  Serial.begin(115200);
  strip.begin();
  strip.show(); // Initialize all pixels to 'off'
  delay(250);
  colorWipe(red, 50);
  Serial.println("Equinox Clock using Neopixels");
  Serial.print("Connecting to ");
  Serial.println(ssid);
  WiFi.begin(ssid, pass);
  WiFi.hostname( "Equinox_Clock" );

  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  colorWipe(green, 50);
  Serial.print("IP number assigned by DHCP is ");
  Serial.println(WiFi.localIP());
  Serial.println("Starting UDP");
  Udp.begin(localPort);
  Serial.print("Local port: ");
  Serial.println(Udp.localPort());
  Serial.println("waiting for sync");
  setSyncProvider(getNtpTime);
  setSyncInterval(3600);
  while (timeStatus() == timeNotSet){
    theaterChase(amber, 50);
  }
}

void loop()
{
  if (timeStatus() != timeNotSet) {
    if (now() != prevDisplay) { //update the display only if time has changed
      prevDisplay = now();
      // the rules for DST changeover are different in New Zealand compared to the UK
      // find the last Sunday in April(4) as the day the time switches to DST
      // start at April 30 and work backwards to find the last Sunday
      int find_start_DST = 30;
      while (find_day_of_the_week(find_start_DST, 4, year())!= "Sunday"){
        find_start_DST-=1;
      }
  
      // find the last Sunday in September (9) as the day the time switches back
      // start at September 30 and work backwards to find the Sunday
      int find_start_GMT = 30;
      while (find_day_of_the_week(find_start_GMT, 9, year())!= "Sunday"){
        find_start_GMT-=1;
      }
      // find the unix timestamp for the switch over dates starting at 2 o'clock in the morning
      long DST_begins = makeTime(2,0,0,find_start_DST,4,year());
      long GMT_begins = makeTime(2,0,0,find_start_GMT,9,year());

      // if the current time is between the above dates, then daylight saving time will be required
      if (now()>DST_begins and now()<GMT_begins){
        DST=1;
      }
      else
      {
        DST=0;
      }

      uint8_t neo_second = second();
      uint8_t neo_minute = minute();
      uint8_t neo_hour = hour() + DST;// add on daylight saving time if necessary
      // digital clock serial display of the time
      Serial.print(neo_hour);
      Serial.print(":");
      if (minute()<10){Serial.print("0");}
      Serial.print(neo_minute);
      Serial.print(":");
      if (second()<10){Serial.print("0");}
      Serial.println(neo_second);

      // calculate how far around the clock the hour hand has moved
      neo_hour = (neo_hour % 12) * 5;
      // move the minute hand slowly around the clock
      if (neo_minute >= 12){neo_hour+=1;} // shift the hour hand around as the minutes change
      if (neo_minute >= 24){neo_hour+=1;}
      if (neo_minute >= 36){neo_hour+=1;}
      if (neo_minute >= 48){neo_hour+=1;}
    
    // switch off all the pixels
    for (uint8_t neo_pixel=0; neo_pixel<61; neo_pixel++){ 
     strip.setPixelColor(neo_pixel, strip.Color(0,0,0)); 
    } 
    // bright white light for the main numbers - no need to shift these 180 degrees as they are symmetrical
//    strip.setPixelColor(0,dim_white); // 12 o'clock position
//    strip.setPixelColor(15,dim_white); // 3 o'clock position
//    strip.setPixelColor(30,dim_white); // 6 o'clock position
//    strip.setPixelColor(45,dim_white); // 9 o'clock position 
    // bright white light for the 5 minute intervals
//    strip.setPixelColor(5,dim_white);
//    strip.setPixelColor(10,dim_white);
//    strip.setPixelColor(20,dim_white);
//    strip.setPixelColor(25,dim_white);
//    strip.setPixelColor(35,dim_white);
//    strip.setPixelColor(40,dim_white);
//    strip.setPixelColor(50,dim_white);
//    strip.setPixelColor(55,dim_white);

    // shift the time leds by 180 degrees, so that the bottom entry
    // represents 6 o'clock
    // add 30 to each value and take the modulus of 60 to shift 180 degrees
    neo_hour += 30; neo_hour = neo_hour % 60;
    neo_minute += 30; neo_minute = neo_minute % 60;
    neo_second += 30; neo_second = neo_second % 60;
    
        
    // choose which leds to light up for the time
    for (uint8_t neo_pixel=0; neo_pixel<61; neo_pixel++){          
      if (neo_pixel == neo_hour){strip.setPixelColor(neo_hour,red);}      
      if (neo_pixel == neo_minute){strip.setPixelColor(neo_minute,blue);}
      if (neo_pixel == neo_second){strip.setPixelColor(neo_second,green);}
    }

    
    
    strip.show();
    // on the hour, do a rainbow effect, with a lightshow at midday
    if (minute() == 0 and second() == 0){
      if (hour() + DST == 12 or hour()+DST == 0){
        colorWipe(bright_white, 50); colorWipe(red, 50); colorWipe(green, 50);
        colorWipe(blue, 50); theaterChase(dim_white, 50); theaterChase(red, 50);
        theaterChase(blue, 50); rainbow(20); rainbowCycle(20); theaterChaseRainbow(50);
        o_clock_chime();
      }
      else{
        theaterChase(bright_white, 50); theaterChase(red, 50); theaterChase(blue, 50);
        theaterChase(green, 50); rainbow(20); rainbowCycle(20); theaterChaseRainbow(50);
        o_clock_chime();
      }
    }
      // on each quarter of an hour, show an effect
      if (second() ==0){
        if (minute() == 15){
          colorWipe(bright_white, 50);
          colorWipe(red, 50);
          colorWipe(green, 50);
          colorWipe(blue, 50);
          quarter_past_chime(); // start and end pixels for quarter past
          }
        if (minute() == 30){
          colorWipe(bright_white, 50);
          theaterChaseRainbow(50);
          half_past_chime();
          }
        if (minute() == 45){
          colorWipe(bright_white, 50); 
          rainbowCycle(20);rainbow(20);
          quarter_to_chime();
          }
      }
    }
  }
}

/*-------- NTP code ----------*/

const int NTP_PACKET_SIZE = 48; // NTP time is in the first 48 bytes of message
byte packetBuffer[NTP_PACKET_SIZE]; //buffer to hold incoming & outgoing packets

time_t getNtpTime()
{
  IPAddress ntpServerIP; // NTP server's ip address

  while (Udp.parsePacket() > 0) ; // discard any previously received packets
  Serial.println("Transmit NTP Request");
  // get a random server from the pool
  WiFi.hostByName(ntpServerName, ntpServerIP);
  Serial.print(ntpServerName);
  Serial.print(": ");
  Serial.println(ntpServerIP);
  sendNTPpacket(ntpServerIP);
  uint32_t beginWait = millis();
  while (millis() - beginWait < 1500) {
    int size = Udp.parsePacket();
    if (size >= NTP_PACKET_SIZE) {
      Serial.println("Receive NTP Response");
      Udp.read(packetBuffer, NTP_PACKET_SIZE);  // read packet into the buffer
      unsigned long secsSince1900;
      // convert four bytes starting at location 40 to a long integer
      secsSince1900 =  (unsigned long)packetBuffer[40] << 24;
      secsSince1900 |= (unsigned long)packetBuffer[41] << 16;
      secsSince1900 |= (unsigned long)packetBuffer[42] << 8;
      secsSince1900 |= (unsigned long)packetBuffer[43];
      return secsSince1900 - 2208988800UL + timeZone * SECS_PER_HOUR;
    }
  }
  Serial.println("No NTP Response :-(");
  return 0; // return 0 if unable to get the time
}

// send an NTP request to the time server at the given address
void sendNTPpacket(IPAddress &address)
{
  Serial.println("Syncing Time Online");
  // set all bytes in the buffer to 0
  memset(packetBuffer, 0, NTP_PACKET_SIZE);
  // Initialize values needed to form NTP request
  // (see URL above for details on the packets)
  packetBuffer[0] = 0b11100011;   // LI, Version, Mode
  packetBuffer[1] = 0;     // Stratum, or type of clock
  packetBuffer[2] = 6;     // Polling Interval
  packetBuffer[3] = 0xEC;  // Peer Clock Precision
  // 8 bytes of zero for Root Delay & Root Dispersion
  packetBuffer[12] = 49;
  packetBuffer[13] = 0x4E;
  packetBuffer[14] = 49;
  packetBuffer[15] = 52;
  // all NTP fields have been given values, now
  // you can send a packet requesting a timestamp:
  Udp.beginPacket(address, 123); //NTP requests are to port 123
  Udp.write(packetBuffer, NTP_PACKET_SIZE);
  Udp.endPacket();
}
